@@ -7,7 +7,9 @@ NULL
 #' adding a custom reporter without removing the current one.
 #'
 #' @export
-MultiReporter <- R6::R6Class("MultiReporter", inherit = Reporter,
+#' @family reporters
+MultiReporter <- R6::R6Class("MultiReporter",
+  inherit = Reporter,
   public = list(
     reporters = list(),
 
@@ -19,6 +21,9 @@ MultiReporter <- R6::R6Class("MultiReporter", inherit = Reporter,
     start_reporter = function() {
       o_apply(self$reporters, "start_reporter")
     },
+    start_file = function(filename) {
+      o_apply(self$reporters, "start_file", filename)
+    },
     start_context = function(context) {
       o_apply(self$reporters, "start_context", context)
     },
@@ -36,6 +41,9 @@ MultiReporter <- R6::R6Class("MultiReporter", inherit = Reporter,
     },
     end_reporter = function() {
       o_apply(self$reporters, "end_reporter")
+    },
+    end_file = function() {
+      o_apply(self$reporters, "end_file")
     }
   )
 )
@@ -43,4 +51,3 @@ MultiReporter <- R6::R6Class("MultiReporter", inherit = Reporter,
 o_apply <- function(objects, method, ...) {
   lapply(objects, function(x) x[[method]](...))
 }
-
